@@ -26,6 +26,20 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
+#ifndef HAVE_SYSCTLBYNAME
+#ifdef HAVE_SYS_PARAM_H
+#include <sys/param.h>
+#endif
+#ifdef HAVE_SYS_MOUNT_H
+#include <sys/mount.h>
+#endif
+#ifdef HAVE_SYS_POLL_H
+#include <sys/poll.h>
+#endif
+#endif
+#ifdef HAVE_POLL_H
+#include <poll.h>
+#endif
 #include <stdio.h>
 #include <unistd.h>
 #include <sys/time.h>
@@ -48,6 +62,10 @@
 #include <sys/mnttab.h>
 #endif
 
+#ifdef HAVE_SYS_VFSTAB_H
+#include <sys/vfstab.h>
+#endif
+
 #if defined(HAVE_SYS_MNTCTL_H) && defined(HAVE_SYS_VMOUNT_H) && defined(HAVE_SYS_VFS_H)
 #include <sys/mntctl.h>
 #include <sys/vfs.h>
@@ -65,6 +83,8 @@
 #endif
 #endif
 
+#define STAT_TIMEOUT_SECONDS 3
+
 
 /* Ideally this should not nonblocking stat, since that can block on
  * downed NFS mounts forever, however there seems to be no good way
@@ -78,8 +98,12 @@ _gnome_vfs_unix_mount_get_unix_device (GnomeVFSUnixMount *mount)
 	dev_t unix_device;
 	pid_t pid;
 	int pipes[2];
-	fd_set read_fds;
+#ifdef HAVE_POLL
+	struct pollfd poll_fd;
+#else
 	struct timeval tv;
+	fd_set read_fds;
+#endif
 	int res;
 	int status;
 
@@ -90,6 +114,8 @@ _gnome_vfs_unix_mount_get_unix_device (GnomeVFSUnixMount *mount)
 	pid = fork ();
 
 	if (pid == -1) {
+		close (pipes[0]);
+		close (pipes[1]);
 		return 0;
 	}
 
@@ -111,12 +137,14 @@ _gnome_vfs_unix_mount_get_unix_device (GnomeVFSUnixMount *mount)
 				write (pipes[1], (char *)&statbuf.st_dev, sizeof (dev_t));
 			}
 		} else {
-			close (pipes[0]);
 			close (pipes[1]);
 		}
 
 		_exit (0);
 	} else {
+		/* Parent */
+		close (pipes[1]);
+		
         wait_again:
 		if (waitpid (pid, &status, 0) < 0) {
 			if (errno == EINTR)
@@ -126,20 +154,25 @@ _gnome_vfs_unix_mount_get_unix_device (GnomeVFSUnixMount *mount)
 			else
 				g_warning ("waitpid() should not fail in gnome_vfs_unix_mount_get_unix_device");
 		}
+
+
+		do {
+#ifdef HAVE_POLL
+			poll_fd.fd = pipes[0];
+			poll_fd.events = POLLIN;
+			res = poll (&poll_fd, 1, STAT_TIMEOUT_SECONDS*1000);
+#else
+			tv.tv_sec = STAT_TIMEOUT_SECONDS;
+			tv.tv_usec = 0;
+			
+			FD_ZERO(&read_fds);
+			FD_SET(pipes[0], &read_fds);
+			
+			res = select (pipes[0] + 1,
+				      &read_fds, NULL, NULL, &tv);
+#endif
+		} while (res == -1 && errno == EINTR);
 		
-		/* Parent */
-		close (pipes[1]);
-
-		FD_ZERO(&read_fds);
-		FD_SET(pipes[0], &read_fds);
-
-		/* Wait max half a second */
-		tv.tv_sec = 0;
-		tv.tv_usec = 500;
-
-		res = select(pipes[0] + 1,
-			     &read_fds, NULL, NULL, &tv);
-
 		if (res > 0) {
 			if (read (pipes[0], (char *)&unix_device, sizeof (dev_t)) != sizeof (dev_t)) {
 				unix_device = 0; 
@@ -195,37 +228,58 @@ _gnome_vfs_get_current_unix_mounts (GList **return_list)
 	char *stat_file;
 	struct stat sb;
 	GnomeVFSUnixMount *mount_entry;
+	GHashTable *mounts_hash;
 	
 	read_file = get_mtab_read_file ();
 	stat_file = get_mtab_monitor_file ();
 
 	*return_list = NULL;
-	
-	if (last_mtime != 0) {
-		if (stat (stat_file, &sb) < 0) {
-			g_warning ("Unable to stat %s: %s", stat_file,
-				   g_strerror (errno));
-			return TRUE;
-		}
-		
-		if (sb.st_mtime == last_mtime) {
-			return FALSE;
-		}
-		last_mtime = sb.st_mtime;
+
+	if (stat (stat_file, &sb) < 0) {
+		g_warning ("Unable to stat %s: %s", stat_file,
+			   g_strerror (errno));
+		return TRUE;
 	}
 
+	if (sb.st_mtime == last_mtime) {
+		return FALSE;
+	}
+
+	last_mtime = sb.st_mtime;
+
 	file = setmntent (read_file, "r");
 	if (file == NULL) {
 		return TRUE;
 	}
 
+	mounts_hash = g_hash_table_new (g_str_hash, g_str_equal);
+	
 	while ((mntent = getmntent (file)) != NULL) {
+		/* ignore any mnt_fsname that is repeated and begins with a '/'
+		 *
+		 * We do this to avoid being fooled by --bind mounts, since
+		 * these have the same device as the location they bind to.
+		 * Its not an ideal solution to the problem, but its likely that
+		 * the most important mountpoint is first and the --bind ones after
+		 * that aren't as important. So it should work.
+		 *
+		 * The '/' is to handle procfs, tmpfs and other no device mounts.
+		 */
+		if (mntent->mnt_fsname != NULL &&
+		    mntent->mnt_fsname[0] == '/' &&
+		    g_hash_table_lookup (mounts_hash, mntent->mnt_fsname)) {
+			continue;
+		}
+
 		mount_entry = g_new0 (GnomeVFSUnixMount, 1);
 
 		mount_entry->mount_path = g_strdup (mntent->mnt_dir);
 		mount_entry->device_path = g_strdup (mntent->mnt_fsname);
 		mount_entry->filesystem_type = g_strdup (mntent->mnt_type);
 		
+		g_hash_table_insert (mounts_hash,
+				     mount_entry->device_path,
+				     mount_entry->device_path);
 
 #if defined (HAVE_HASMNTOPT)
 		if (hasmntopt (mntent, MNTOPT_RO) != NULL) {
@@ -234,6 +288,7 @@ _gnome_vfs_get_current_unix_mounts (GList **return_list)
 #endif
 		*return_list = g_list_prepend (*return_list, mount_entry);
 	}
+	g_hash_table_destroy (mounts_hash);
 	
 	endmntent (file);
 	
@@ -275,19 +330,19 @@ _gnome_vfs_get_current_unix_mounts (GList **return_list)
 	stat_file = get_mtab_monitor_file ();
 
 	*return_list = NULL;
-	
-	if (last_mtime != 0) {
-		if (stat (stat_file, &sb) < 0) {
-			g_warning ("Unable to stat %s: %s", stat_file,
-				   g_strerror (errno));
-			return TRUE;
-		}
-		
-		if (sb.st_mtime == last_mtime) {
-			return FALSE;
-		}
+
+	if (stat (stat_file, &sb) < 0) {
+		g_warning ("Unable to stat %s: %s", stat_file,
+			   g_strerror (errno));
+		return TRUE;
 	}
 
+	if (sb.st_mtime == last_mtime) {
+		return FALSE;
+	}
+
+	last_mtime = sb.st_mtime;
+
 	file = setmntent (read_file, "r");
 	if (file == NULL) {
 		return TRUE;
@@ -420,7 +475,7 @@ _gnome_vfs_get_current_unix_mounts (GList **return_list)
 		mount_entry->mount_path = g_strdup (mntent[i].f_mntonname);
 		mount_entry->device_path = g_strdup (mntent[i].f_mntfromname);
 		mount_entry->filesystem_type = g_strdup (mntent[i].f_fstypename);
-		if (mntent[i].f_flags == MNT_RDONLY) {
+		if (mntent[i].f_flags & MNT_RDONLY) {
 		    	mount_entry->is_read_only = TRUE;
 		}
 
@@ -446,7 +501,9 @@ get_fstab_file (void)
 {
 #ifdef _PATH_MNTTAB
 	return _PATH_MNTTAB;
-#else	
+#elif defined VFSTAB
+	return VFSTAB;
+#else
 	return "/etc/fstab";
 #endif
 }
@@ -467,19 +524,19 @@ _gnome_vfs_get_unix_mount_table (GList **return_list)
 	stat_file = read_file = get_fstab_file ();
 
 	*return_list = NULL;
-	
-	if (last_mtime != 0) {
-		if (stat (stat_file, &sb) < 0) {
-			g_warning ("Unable to stat %s: %s", stat_file,
-				   g_strerror (errno));
-			return TRUE;
-		}
-		
-		if (sb.st_mtime == last_mtime) {
-			return FALSE;
-		}
+
+	if (stat (stat_file, &sb) < 0) {
+		g_warning ("Unable to stat %s: %s", stat_file,
+			   g_strerror (errno));
+		return TRUE;
 	}
 
+	if (sb.st_mtime == last_mtime) {
+		return FALSE;
+	}
+
+	last_mtime = sb.st_mtime;
+
 	file = setmntent (read_file, "r");
 	if (file == NULL) {
 		return TRUE;
@@ -517,7 +574,8 @@ _gnome_vfs_get_unix_mount_table (GList **return_list)
 
 		if ((mntent->mnt_type != NULL && strcmp ("supermount", mntent->mnt_type) == 0)
 #ifdef HAVE_HASMNTOPT
-		    || hasmntopt (mntent, "user") != NULL
+		    || (hasmntopt (mntent, "user") != NULL
+			&& hasmntopt (mntent, "user") != hasmntopt (mntent, "user_xattr"))
 		    || hasmntopt (mntent, "users") != NULL
 		    || hasmntopt (mntent, "owner") != NULL
 #endif
@@ -552,19 +610,19 @@ _gnome_vfs_get_unix_mount_table (GList **return_list)
 	stat_file = read_file = get_fstab_file ();
 
 	*return_list = NULL;
-	
-	if (last_mtime != 0) {
-		if (stat (stat_file, &sb) < 0) {
-			g_warning ("Unable to stat %s: %s", stat_file,
-				   g_strerror (errno));
-			return TRUE;
-		}
-		
-		if (sb.st_mtime == last_mtime) {
-			return FALSE;
-		}
+
+	if (stat (stat_file, &sb) < 0) {
+		g_warning ("Unable to stat %s: %s", stat_file,
+			   g_strerror (errno));
+		return TRUE;
 	}
 
+	if (sb.st_mtime == last_mtime) {
+		return FALSE;
+	}
+
+	last_mtime = sb.st_mtime;
+
 	file = setmntent (read_file, "r");
 	if (file == NULL) {
 		return TRUE;
@@ -594,7 +652,8 @@ _gnome_vfs_get_unix_mount_table (GList **return_list)
 
 		if ((mntent.mnt_fstype != NULL)
 #ifdef HAVE_HASMNTOPT
-		    || hasmntopt (&mntent, "user") != NULL
+		    || (hasmntopt (&mntent, "user") != NULL
+			&& hasmntopt (&mntent, "user") != hasmntopt (&mntent, "user_xattr"))
 		    || hasmntopt (&mntent, "users") != NULL
 		    || hasmntopt (&mntent, "owner") != NULL
 #endif
@@ -815,7 +874,17 @@ _gnome_vfs_get_unix_mount_table (GList **return_list)
 	}
 
 #ifdef HAVE_SYS_SYSCTL_H
+#if defined(HAVE_SYSCTLBYNAME)
 	sysctlbyname ("vfs.usermount", &usermnt, &len, NULL, 0);
+#elif defined(CTL_VFS) && defined(VFS_USERMOUNT)
+	{
+		int mib[2];
+
+		mib[0] = CTL_VFS;
+		mib[1] = VFS_USERMOUNT;
+		sysctl (mib, 2, &usermnt, &len, NULL, 0);
+	}
+#endif
 #endif
 
 	while ((fstab = getfsent ()) != NULL) {
@@ -835,8 +904,9 @@ _gnome_vfs_get_unix_mount_table (GList **return_list)
 
 #ifdef HAVE_SYS_SYSCTL_H
 		if (usermnt != 0) {
+			uid_t uid = getuid ();
 			if (stat (fstab->fs_file, &sb) == 0) {
-				if (sb.st_uid != 0) {
+				if (uid == 0 || sb.st_uid == uid) {
 					mount_entry->is_user_mountable = TRUE;
 				}
 			}
@@ -861,8 +931,8 @@ static GnomeVFSMonitorHandle *fstab_monitor = NULL;
 static GnomeVFSUnixMountCallback fstab_callback = NULL;
 static GnomeVFSMonitorHandle *mtab_monitor = NULL;
 static GnomeVFSUnixMountCallback mtab_callback = NULL;
-guint mtab_poll_tag = 0;
-guint fstab_poll_tag = 0;
+static guint mtab_poll_tag = 0;
+static guint fstab_poll_tag = 0;
 
 static void
 fstab_monitor_callback (GnomeVFSMonitorHandle *handle,
