@@ -315,6 +315,8 @@ _gnome_vfs_pty_run_on_pty(int fd,
 		fprintf(stderr, "Child sending child-ready.\n");
 	}
 #endif
+	/* Initialize so valgrind doesn't complain */
+	c = 0;
 	n_write(ready_writer, &c, 1);
 	fsync(ready_writer);
 	n_read(ready_reader, &c, 1);
@@ -607,6 +609,8 @@ _gnome_vfs_pty_getpt(void)
 #ifdef HAVE_GETPT
 	/* Call the system's function for allocating a pty. */
 	fd = getpt();
+#elif defined(HAVE_POSIX_OPENPT)
+	fd = posix_openpt(O_RDWR | O_NOCTTY);
 #else
 	/* Try to allocate a pty by accessing the pty master multiplex. */
 	fd = open("/dev/ptmx", O_RDWR | O_NOCTTY);
